// partition:  a container for organizing a set of sequential integers
//             into non-overlapping groups.
// Version:    0.1
// Copyright 2013 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

package org.mattbusche.util.partition;

/** <p>Maintains a doubly linked list of Entry&lt;M&gt; objects from a single
 ** associated Domain and may also be assigned a user-defined value of type
 ** V.</p>
 **
 ** <p>Read the <a href="package-summary.html">package summary</a> for
 ** introductory information about Group and its related classes.</p>
 **
 ** <p>An Entry may be added to a Group via add operations typical of linked
 ** lists, but each such add has the atypical side-affect of simultaneously
 ** removing the Entry from its previous Group membership (if any) all with
 ** constant computational complexity.</p>
 **
 ** <p>The following four terms are used consistently throughout to refer
 ** to key list positions:</p>
 ** <ul>
 ** <li><i>beforeFront</i> - One position before the first entry in the list.</li>
 ** <li><i>front</i> - The first entry in the list.</li>
 ** <li><i>back</i> - The last entry in the list.</li>
 ** <li><i>afterBack</i> - One position after the last entry in the list.</li>
 ** </ul>
 **/
public class Group<M,V>
{
    Domain<M,V> _domain;
    Node<M,V> _end;
    int _size;
    V _value;

    /** Constructs an empty <code>Group</code> with no {@link Domain}
     ** association and a <code>null</code> value.  Few operations on a Group
     ** constructed in this way are permitted until a <code>Domain</code>
     ** object has been associated with the <code>Group</code> via the
     ** {@link #setDomain} method.
     **/
    public Group()
    {
    }


    /** Constructs an empty <code>Group</code> associated with {@link Domain}
     ** <code>d</code> and sets the <code>Group</code>'s value to
     ** <code>null</code>.
     **/
    public Group(Domain<M,V> d)
    {
        _domain = d;
        _end = new Node<M,V>(this);
    }


    /** Constructs an empty <code>Group</code> associated with {@link Domain}
     ** <code>d</code> and sets the <code>Group</code>'s value to
     ** <code>v</code>.
     **/
    public Group(Domain<M,V> d, V v)
    {
        this(d);
        _value = v;
    }


    /** Constructs a <code>Group</code> that is a copy of <code>g</code> with the exception that
     ** its associated with {@link Domain} <code>d</code>.
     ** <code>Domain</code> <code>d</code> must not be the same
     ** <code>Domain</code> as that associated with <code>g</code>.
     **
     ** @throws NullPointerException if <code>d</code> is <code>null</code>.
     **
     ** @throws PartitionException if <code>g</code> is associated with
     ** <code>Domain</code> <code>d</code>.
     **
     ** @throws PartitionException if an {@link Entry} of <code>g</code> has
     ** an ID not contained in <code>d</code>.
     **/
    public Group(Domain<M,V> d, Group<M,V> g)
    {
        this(d);

        // Check for d being null explicitly; otherwise, the case
        // where all of the following are true would not throw (which
        // would just be weird and difficult to document):
        //
        //    o d == null
        //    o g.getDomain() != null
        //    o g.size() == 0.
        //
        if(d == null)
            throw new NullPointerException("Domain argument cannot be null.");

        _value = g.getValue();

        if(this._domain == g._domain)
            throw new PartitionException("Attempt to copy a Group to another Group in the same Domain.");

        for(GroupIterator<M,V> i = g.front(); !i.isAfterBack(); i.inc())
            addBack(i.getEntry().getId());
    }


    /** Returns the {@link Domain} associated with this <code>Group</code> (or
     ** <code>null</code> if this <code>Group</code> has no
     ** <code>Domain</code> association).
     **/
    public Domain<M,V> getDomain()
    {
        return _domain;
    }



    /** Associates this <code>Group</code> with <code>d</code>.  If this
     ** <code>Group</code> was not previously associated with <code>d</code>,
     ** then the <code>Group</code>'s list is first cleared of any {@link
     ** Entry} objects from its current {@link Domain} association (if any).
     ** If this <code>Group</code> was already associated with <code>d</code>,
     ** then there is no effect.
     **/
    public void setDomain(Domain<M,V> d)
    {
        if(_domain == d)
            return;
        removeAll();
        _domain = d;
    }


    /** Returns the user provided value object associated with this Group.
     **/
    public V getValue()
    {
        return _value;
    }


    /** Sets <code>v</code> as this <code>Group</code>'s value.
     **/
    public void setValue(V v)
    {
        this._value = v;
    }


    /** Returns the number of {@link Entry} objects in this
     ** <code>Group</code>'s list.
     **/
    public int size()
    {
        return _size;
    }


    /** Returns <code>true</code> if this <code>Group</code> contains the
     ** {@link Entry} with the given <code>id</code> and <code>false</code>
     ** otherwise.
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws NullPointerException if this <code>Group</code> has no {@link
     ** Domain} association.
     ** @throws PartitionException if <code>id</code> is not a member of the
     ** associated <code>Domain</code>.
     **/
    public boolean contains(int id)
    {
        return _domain._getNodeChecked(id)._group == this;
    }


    /** Returns the {@link Entry} object at the front of the
     ** <code>Group</code> list while leaving the list unmodified.
     **
     ** @throws PartitionException if the Group is empty.
     **/
    public Entry<M> peekFront()
    {
        _verifyNotEmpty();
        return _end._next._entry;
    }


    /** Returns the {@link Entry} object at the back of the <code>Group</code>
     ** list while leaving the list unmodified.
     **
     ** @throws PartitionException if the <code>Group</code> is empty.
     **/
    public Entry<M> peekBack()
    {
        _verifyNotEmpty();
        return _end._prev._entry;
    }


    /** <p>Adds the {@link Entry} with the given <code>id</code> from the
     ** <code>Group</code>'s associated {@link Domain} to the front of this
     ** <code>Group</code>'s list.  If the <code>Entry</code> was previously a
     ** member of a <code>Group</code>, it is first removed from that
     ** <code>Group</code>.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws NullPointerException if this <code>Group</code> has no
     ** <code>Domain</code> association.
     **
     ** @throws PartitionException if <code>id</code> is not a member of the
     ** associated <code>Domain</code>.
     **/
    public void addFront(int id)
    {
        _insertAfter(_end, id);
    }


    /** <p>Adds the {@link Entry} with the given <code>id</code> from the
     ** <code>Group</code>'s associated {@link Domain} to the back of this
     ** <code>Group</code>'s list.  If the <code>Entry</code> was previously a
     ** member of a <code>Group</code>, it is first removed from that
     ** <code>Group</code>.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws NullPointerException if this <code>Group</code> has no
     ** <code>Domain</code> association.
     **
     ** @throws PartitionException if <code>id</code> is not a member of the
     ** associated <code>Domain</code>.
     **/
    public void addBack(int id)
    {
        _insertAfter(_end._prev, id);
    }


    /** <p>Moves all {@link Entry} objects from <code>Group</code>
     ** <code>g</code> to the front of this <code>Group</code>'s list.</p>
     **
     ** <p>Computational Complexity:  Linear in the size of <code>g</code>.</p>
     **
     ** @throws PartitionException if <code>Group</code> <code>g</code> is not
     ** associated with the same {@link Domain} as this <code>Group</code>.
     **/
    public void addFront(Group<M,V> g)
    {
        if(this == g)
            return;

        if(this._domain != g._domain)
            throw new PartitionException("Attempt to push a Group to another Group in a different Domain.");

        if(g._size == 0)
            return;

        Node<M,V> na = _end;
        Node<M,V> nb = g._end._next;
        Node<M,V> nc = g._end._prev;
        Node<M,V> nd = _end._next;

        GroupIterator<M,V> i = new GroupIterator<M,V>(this, nd);

        na._next = nb;
        nb._prev = na;
        nc._next = nd;
        nd._prev = nc;

        g._end._next = g._end;
        g._end._prev = g._end;
        _size += g._size;
        g._size = 0;

        // Still need to update all the group references.
        //
        while(!i.dec().isBeforeFront())
            i._node._group = this;
    }


    /** <p>Moves all {@link Entry} objects from <code>g</code> to the back of
     ** this <code>Group</code>'s <code>Entry</code> list.</p>
     **
     ** <p>Computational Complexity:  Linear in the size of <code>g</code>.</p>
     **
     ** @throws PartitionException if Group <code>g</code> is not associated
     ** with the same <code>Domain</code> as this <code>Group</code>.
     **/
    public void addBack(Group<M,V> g)
    {
        if(this == g)
            return;

        if(this._domain != g._domain)
            throw new PartitionException("Attempt to push a Group to another Group in a different Domain.");

        if(g._size == 0)
            return;

        GroupIterator<M,V> i = new GroupIterator<M,V>(this, _end._prev);

        Node<M,V> na = _end._prev;
        Node<M,V> nb = g._end._next;
        Node<M,V> nc = g._end._prev;
        Node<M,V> nd = _end;

        na._next = nb;
        nb._prev = na;
        nc._next = nd;
        nd._prev = nc;

        g._end._next = g._end;
        g._end._prev = g._end;

        _size += g._size;
        g._size = 0;

        // Still need to update all the group references.
        //
        while(!i.inc().isAfterBack())
            i._node._group = this;
    }


    /** <p>Adds all {@link Entry} objects from this <code>Group</code>'s
     ** {@link Domain} to this <code>Group</code>'s list (simultaneously
     ** removing them from the <code>Group</code>s with which they were
     ** previously a member).  <code>Entry</code> objects are listed in
     ** ascending ID order.</p>
     **
     ** <p>Computational Complexity:  Linear in the size of this
     ** <code>Group</code>'s <code>Domain</code>.</p>
     **
     ** @throws NullPointerException if this Group has no Domain association.
     **/
    public void addAll()
    {
        // An optimization idea:
        //    1. return if group size == domain size
        //    2. for n in domain nodes: if(n->_group) n->_group->removeAll()
        //    3. for n in domain nodes: build linked-list on this group.
        //
        int begin = _domain.min();
        int end = _domain.min() + _domain.size();
        for(int id = begin; id < end; ++id)
            addBack(id);
    }


    /** <p>Removes the {@link Entry} at the front of this <code>Group</code>'s
     ** list and returns it.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     ** 
     ** @throws PartitionException if this <code>Group</code> is empty.
     **/
    public Entry<M> removeFront()
    {
        _verifyNotEmpty();
        return _remove(_end._next)._entry;
    }


    /** <p>Removes the {@link Entry} at the back of this Group's list and
     ** returns it.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     ** 
     ** @throws PartitionException if this <code>Group</code> is empty.
     **/
    public Entry<M> removeBack()
    {
        _verifyNotEmpty();
        return _remove(_end._prev)._entry;
    }


    /** <p>Removes the {@link Entry} with the given <code>id</code> from this
     ** <code>Group</code>'s list and returns it.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     ** 
     ** @throws PartitionException if this <code>Group</code> does not contain
     ** the <code>Entry</code> with the given <code>id</code>.
     ** @throws NullPointerException if this <code>Group</code> has no
     ** <code>Domain</code> association.
     **/
    public Entry<M> remove(int id)
    {
        Node<M,V> n = _domain._getNodeChecked(id);
        if(n._group != this)
            throw new PartitionException("ID not in this Group.");
        _remove(n);
        return n._entry;
    }


    /** <p>Removes all <code>Entry</code> objects from this
     ** <code>Group</code>'s list.</p>
     **
     ** <p>Computational Complexity:  Linear in the size of this
     ** <code>Group</code>.</p>
     **/
    public void removeAll()
    {
        // Could just write:
        //
        //   while(size())
        //     remove_front();
        //
        // But this should be considerably faster:
        //
        Node<M,V> n = _end._next;
        while(n != _end)
        {
            Node<M,V> tmp = n;
            n = n._next;
            tmp._group = null;

            // Nulling these out for fear it will confuse the GC.
            //
            tmp._prev = null;
            tmp._next = null;
        }

        _end._next = _end;
        _end._prev = _end;
        _size      = 0;
    }


    /** <p>Returns a {@link GroupIterator} pointing to the first
     ** {@link Entry} of this <code>Group</code>'s list.  If this
     ** <code>Group</code> is empty, the returned <code>GroupIterator</code>
     ** will point to the {@link #afterBack} list position.</p>
     **/
    public GroupIterator<M,V> front()
    {
        return new GroupIterator<M,V>(this, _end._next);
    }


    /** <p>Returns a {@link GroupIterator} pointing to the last {@link Entry}
     ** of this <code>Group</code>'s list.  If this <code>Group</code> is
     ** empty, the returned <code>GroupIterator</code> will point to the
     ** {@link #beforeFront} list position.</p>
     **/
    public GroupIterator<M,V> back()
    {
        return new GroupIterator<M,V>(this, _end._prev);
    }


    /** <p>Returns a {@link GroupIterator} pointing to the position before the
     ** first {@link Entry} in this <code>Group</code>'s list.</p>
     **/
    public GroupIterator<M,V> beforeFront()
    {
        return new GroupIterator<M,V>(this, _end);
    }


    /** <p>Returns a {@link GroupIterator} pointing to the position after the
     ** last {@link Entry} in this <code>Group</code>'s list.</p>
     **/
    public GroupIterator<M,V> afterBack()
    {
        return new GroupIterator<M,V>(this, _end);
    }


    /** <p>Returns a {@link GroupIterator} pointing to the {@link Entry} from
     ** this list with the given <code>id</code>, or pointing to the {@link
     ** #afterBack} position if this <code>Group</code> does not contain the
     ** <code>Entry</code> with the given <code>id</code>.</p>
     **
     ** <p><em>PLEASE NOTE THE EXCEPTIONAL CONDITION BELOW.</em></p>
     **
     ** @throws PartitionException if <code>id</code> is not a member of the
     ** associated <code>Domain</code>.
     **/
    public GroupIterator<M,V> find(int id)
    {
        Node<M,V> n = _domain._getNodeChecked(id);
        return new GroupIterator<M,V>(this, n._group == this ? n : _end);
    }


    /** <p>Prints the Group in the following format:</p>
     ** <pre>
     **       GroupValue:[Entry_1, Entry_2, ..., Entry_N]
     ** </pre>
     ** <p>where each Entry is printed as <code>(id, member)</code>.</p>
     **/
    @Override
    public String toString()
    {
        StringBuilder sb = new StringBuilder();
        sb.append(_value);
        sb.append(":[");
        boolean first = true;
        for(GroupIterator<M,V> i = front(); !i.isAfterBack(); i.inc()) {
            if(first)
                first = false;
            else
                sb.append(", ");
            sb.append(i.getEntry());
        }
        sb.append("]");
        return sb.toString();
    }


    void _insertAfter(Node<M,V> n1, int id)
    {
        Node<M,V> n3 = n1._next;
        Node<M,V> n2 = _domain._getNodeChecked(id);
        if(n2 == n1 || n2 == n3)
            return;
        if(n2._group != null)
            n2._group._remove(n2);
        n1._next = n2;
        n2._prev = n1;
        n2._next = n3;
        n3._prev = n2;
        n2._group = this;
        ++_size;
    }


    Node<M,V> _remove(Node<M,V> node)
    {
        node._prev._next = node._next;
        node._next._prev = node._prev;
        --_size;
        node._group = null;
        node._prev = null;
        node._next = null;
        return node;
    }


    private void _verifyNotEmpty()
    {
        if(_size == 0)
            throw new PartitionException("Group is empty.");
    }
}
